import pytest
import shutil
import hashlib
import tempfile
import itertools
import multiprocessing

from fsdicts import *


@pytest.fixture(params=itertools.product([LinkStorage, ReferenceStorage], [hashlib.md5, hashlib.sha1, hashlib.sha256, hashlib.sha512], [LocalLock, FileLock]))
def storage(request):
    # Untuple the parameters
    storage_type, hash_type, lock_type = request.param

    # Create the storage
    return storage_type(tempfile.mktemp(), hash=hash_type, lock=lock_type)


def test_put(storage):
    # Make sure storage is empty
    assert not len(storage)

    # Add object to storage
    storage.put(b"Hello")

    # Make sure object was registered
    assert len(storage)


def test_purge(storage):
    # Add object to storage
    storage.put(b"Hello World")

    # Purge the database
    storage.purge()

    # Make sure object was deleted
    assert not len(storage)


def test_usage_purge(storage):
    # Create target reference path
    reference_path = tempfile.mktemp()

    # Add a value to the storage
    identifier = storage.put(b"Hello World!!!")

    # Link the identifier
    storage.link(identifier, reference_path)

    # Purge the storage
    storage.purge()

    # Make sure object was not deleted
    assert len(storage)

    # Unlink the identifier
    storage.unlink(reference_path)

    # Make sure object was deleted
    assert not len(storage)


def test_readlink_purge(storage):
    # Create the value to write
    value = b"Hello World!!!"

    # Create target reference path
    reference_path = tempfile.mktemp()

    # Add a value to the storage
    identifier = storage.put(value)

    # Link the identifier
    storage.link(identifier, reference_path)

    # Purge the storage
    storage.purge()

    # Read the link and compare to value
    assert storage.readlink(reference_path) == value

    # Copy the reference path to a tempfile
    fake_reference_path = tempfile.mktemp()
    shutil.copy(reference_path, fake_reference_path)

    # Unlink the reference
    storage.unlink(reference_path)

    # Try reading the original reference
    with pytest.raises((OSError, IOError)):
        storage.readlink(reference_path)


def test_multiprocess_writes(storage):
    # Create global things
    manager = multiprocessing.Manager()
    exceptions = manager.list()

    def stress():
        for _ in range(100):
            try:
                # Create random data
                content = b"A" * 1024 * 1024

                # Write to database
                link = tempfile.mktemp()
                identifier = storage.put(content, link)
                storage.unlink(link)
            except BaseException as e:
                # Append failure
                exceptions.append(e)

    # Create many stress processes
    processes = [multiprocessing.Process(target=stress) for _ in range(10)]

    # Execute all processes
    for p in processes:
        p.start()

    # Wait for all processes
    for p in processes:
        p.join()

    # Raise all of the exceptions
    for e in exceptions:
        raise e



import json

# Create default encoding
ENCODING = "utf-8"

# Create encoder tuples
JSON = (lambda value: json.dumps(value).encode(ENCODING), json.loads)
PYTHON = (lambda value: repr(value).encode(ENCODING), eval)


import pytest
import datetime

from fsdicts import *


@pytest.mark.parametrize("encoder", [PYTHON, JSON])
@pytest.mark.parametrize("value", ["Hello", 10, True, None, {"Hello": "World"}, ["Hello", "World"]])
def test_generic_encode_decode(encoder, value):
    # Untuple encoder
    encode, decode = encoder

    # Try encoding the value
    encoded_value = encode(value)

    # Make sure the encoded value is of type bytes
    assert isinstance(encoded_value, bytes)

    # Decode the value
    decoded_value = decode(encoded_value)

    # Make sure the decoded value equals the original value
    assert decoded_value == value


@pytest.mark.parametrize("value", [b"Hello", "World", ("Hello", "World"), datetime.datetime.now()])
def test_python_encode_decode(value):
    test_generic_encode_decode(PYTHON, value)



from fsdicts.mapping import MutableMapping, Mapping


class AttributeMapping(Mapping):

    def __getattr__(self, key):
        try:
            return object.__getattribute__(self, key)
        except AttributeError:
            # Key is not in prototype chain, try returning
            try:
                return self[key]
            except KeyError:
                # Replace KeyErrors with AttributeErrors
                raise AttributeError(key)


class MutableAttributeMapping(MutableMapping, AttributeMapping):

    def __setattr__(self, key, value):
        try:
            object.__getattribute__(self, key)
        except AttributeError:
            # Convert value to bunch
            if isinstance(value, Mapping):
                value = Bunch(value)

            # Set the item
            self[key] = value
        else:
            # Key is in prototype chain, set it
            object.__setattr__(self, key, value)

    def __delattr__(self, key):
        try:
            object.__getattribute__(self, key)
        except AttributeError:
            # Delete the item
            try:
                del self[key]
            except KeyError:
                # Replace KeyErrors with AttributeErrors
                raise AttributeError(key)
        else:
            # Key is in prototype chain, delete it
            object.__delattr__(self, key)


class Bunch(dict, MutableAttributeMapping):
    pass



import os
import hashlib
import binascii

from fsdicts.lock import LocalLock, FileLock


class Storage(object):

    def __init__(self, path, lock):
        raise NotImplementedError()

    def put(self, value, link=None):
        raise NotImplementedError()

    def readlink(self, link):
        raise NotImplementedError()

    def link(self, identifier, link):
        raise NotImplementedError()

    def unlink(self, link):
        raise NotImplementedError()

    def release(self, identifier):
        raise NotImplementedError()

    def purge(self):
        raise NotImplementedError()

    def __len__(self):
        raise NotImplementedError()


class LinkStorage(Storage):

    def __init__(self, path, lock=LocalLock, hash=hashlib.md5):
        # Make sure the operating system is supported
        if os.name != "posix":
            raise NotImplementedError("Unsupported operating system")

        # Make the path absolute
        path = os.path.abspath(path)

        # Intialize the path and hash
        self._path = path
        self._lock = lock
        self._hash = hash

        # Create the path if needed
        if not os.path.exists(self._path):
            os.makedirs(self._path)

    def _internal_link(self, identifier, link):
        # Link the identifier and the link
        os.link(identifier, link)

    def _internal_release(self, identifier):
        # Make sure the path exists
        if not os.path.isfile(identifier):
            raise ValueError(identifier)

        # If more then one link exists, skip
        if os.stat(identifier).st_nlink > 1:
            return

        # Remove the file
        os.remove(identifier)

    def put(self, value, link=None):
        # Create the value hash
        hash = self._hash(value).hexdigest()

        # Create the object path
        path = os.path.join(self._path, hash)

        # Lock the path
        with self._lock(path):
            # Write file if missing
            if not os.path.isfile(path):
                # Create temporary path for writing
                temporary = path + "." + binascii.b2a_hex(os.urandom(4)).decode()

                # Write the object
                with open(temporary, "wb") as object:
                    object.write(value)

                # Rename the temporary file
                os.rename(temporary, path)

            # Link as needed
            if link:
                self._internal_link(path, link)

        # Return the hash
        return path

    def readlink(self, link):
        # Read the identifier as a file
        with open(link, "rb") as object:
            value = object.read()

        # Return the value
        return value

    def link(self, identifier, link):
        # Lock the identifier
        with self._lock(identifier):
            # Link the identifier and the link
            return self._internal_link(identifier, link)

    def unlink(self, link):
        # Check whether the path exists
        if not os.path.isfile(link):
            raise ValueError(link)

        # Read the link and create hash
        with open(link, "rb") as object:
            object_value = object.read()

        # Create object hash
        object_hash = self._hash(object_value).hexdigest()

        # Create object path
        identifier = os.path.join(self._path, object_hash)

        # Lock the identifier
        with self._lock(identifier):
            # Remove the link
            os.unlink(link)

            # Clean the identifier
            self._internal_release(identifier)

    def release(self, identifier):
        # Lock the identifier
        with self._lock(identifier):
            return self._internal_release(identifier)

    def purge(self):
        # List all files in the storage and check the link count
        for hash in self:
            # Create the file path
            identifier = os.path.join(self._path, hash)

            # Lock the identifier
            with self._lock(identifier):
                # If path does not exist, skip
                if not os.path.isfile(identifier):
                    continue

                # Clean the file
                self._internal_release(identifier)

    def __iter__(self):
        # Calculate the checksum length
        length = len(self._hash().hexdigest())

        # Loop over object directory
        for hash in os.listdir(self._path):
            # Make sure hash matches expected length
            if len(hash) != length:
                continue

            # Yield the hash
            yield hash

    def __len__(self):
        # Count the object files
        return len(list(iter(self)))


class ReferenceStorage(Storage):

    def __init__(self, path, lock=FileLock, hash=hashlib.md5):
        # Make the path absolute
        path = os.path.abspath(path)

        # Intialize the path and hash
        self._lock = lock
        self._hash = hash

        # Create the objects path and references path
        self._objects_path = os.path.join(path, "objects")
        self._references_path = os.path.join(path, "references")

        # Create the objects directory if needed
        if not os.path.isdir(self._objects_path):
            os.makedirs(self._objects_path)

        # Create the references directory if needed
        if not os.path.isdir(self._references_path):
            os.makedirs(self._references_path)

    def _internal_link(self, hash, link):
        # Create the object path
        object_path = os.path.join(self._objects_path, hash)

        # Make sure the object exists
        if not os.path.isfile(object_path):
            raise KeyError(hash)

        # Create references file path
        references_path = os.path.join(self._references_path, hash)

        # Append the path to the file
        with open(references_path, "a") as references_file:
            references_file.write(link + "\n")

        # Write the link file
        with open(link, "w") as file:
            file.write(hash)

    def _internal_release(self, hash):
        # Create the object path
        object_path = os.path.join(self._objects_path, hash)

        # Make sure the object exists
        if not os.path.isfile(object_path):
            raise KeyError(hash)

        # Create references file path
        references_path = os.path.join(self._references_path, hash)

        # If the references path exists, check references
        if os.path.isfile(references_path):
            # Read the list of references from the file
            with open(references_path, "r") as references_file:
                references = references_file.read().splitlines()

            # Filter out references that do not exist
            references = list(filter(lambda reference: os.path.isfile(reference), references))

            # If there are references, return - can't purge
            if references:
                return

            # Remove the references file
            os.remove(references_path)

        # Remove the object path
        os.remove(object_path)

    def put(self, value, link=None):
        # Create the value hash
        hash = self._hash(value).hexdigest()

        # Create the object path
        object_path = os.path.join(self._objects_path, hash)

        # Lock the object
        with self._lock(object_path):
            # Check whether the path exists
            if not os.path.isfile(object_path):
                # Create temporary path for writing
                temporary = object_path + "." + binascii.b2a_hex(os.urandom(4)).decode()

                # Write the object
                with open(temporary, "wb") as object:
                    object.write(value)

                # Rename the temporary file
                os.rename(temporary, object_path)

            # Link if needed
            if link:
                self._internal_link(hash, link)

        # Return the hash
        return hash

    def readlink(self, link):
        # Read the link file
        with open(link, "r") as file:
            hash = file.read()

        # Create the object path
        object_path = os.path.join(self._objects_path, hash)

        # Make sure the object exists
        if not os.path.isfile(object_path):
            raise KeyError(hash)

        # Read the object file
        with open(object_path, "rb") as object:
            value = object.read()

        # Return the value
        return value

    def link(self, hash, link):
        # Create the object path
        object_path = os.path.join(self._objects_path, hash)

        # Lock the object
        with self._lock(object_path):
            return self._internal_link(hash, link)

    def unlink(self, link):
        # Read the link file
        with open(link, "r") as file:
            hash = file.read()

        # Create the object path
        object_path = os.path.join(self._objects_path, hash)

        # Make sure the object exists
        if not os.path.isfile(object_path):
            raise KeyError(hash)

        # Lock the object
        with self._lock(object_path):
            # Create references file path
            references_path = os.path.join(self._references_path, hash)

            # Read the list of references from the file
            with open(references_path, "r") as references_file:
                references = references_file.read().splitlines()

            # Filter out the current reference
            references = list(filter(lambda reference: reference != link, references))

            # Write the list of references back to the file
            with open(references_path, "w") as references_file:
                for reference in references:
                    references_file.write(reference + "\n")

            # Remove the link path
            os.remove(link)

            # Release the object
            self._internal_release(hash)

    def release(self, hash):
        # Create the object path
        object_path = os.path.join(self._objects_path, hash)

        # Lock the object path
        with self._lock(object_path):
            return self._internal_release(hash)

    def purge(self):
        # List all objects in storage
        for hash in self:
            # Release the object
            self.release(hash)

    def __iter__(self):
        # Calculate the checksum length
        length = len(self._hash().hexdigest())

        # Loop over object directory
        for hash in os.listdir(self._objects_path):
            # Make sure hash matches expected length
            if len(hash) != length:
                continue

            # Yield the hash
            yield hash

    def __len__(self):
        # Count the object files
        return len(list(iter(self)))



import os
import time
import random
import hashlib
import tempfile
import threading

from fsdicts.encoders import ENCODING


class FileLock(object):

    def __init__(self, path):
        # Create the lock path
        self._path = path + ".lock"

        # Create internal state
        self._locked = False

    def _try_acquire(self):
        try:
            # Try creating the file
            os.mkdir(self._path)

            # Update lock state
            self._locked = True

            # Locking succeeded
            return True
        except OSError:
            # Locking failed
            return False

    def locked(self):
        return self._locked

    def acquire(self, blocking=True, timeout=None):
        # Mark start time
        start_time = time.time()

        # Try acquiring for the first time
        if self._try_acquire():
            return True

        # If non-blocking, return here
        if not blocking:
            return False

        # Loop until timeout is reached
        while (timeout is None) or (time.time() - start_time) < timeout:
            # Try aquiring the lock
            if self._try_acquire():
                return True

            # Sleep random amount
            time.sleep(random.random() / 1000.0)

        # Timeout reached
        return False

    def release(self):
        # Make sure not already unlocked
        if not self._locked:
            return

        # Try removing the directory
        os.rmdir(self._path)

        # Update the lock status
        self._locked = False

    def __enter__(self):
        # Lock the lock
        self.acquire()

        # Return "self"
        return self

    def __exit__(self, *exc_info):
        # Unlock the lock
        self.release()

    def __str__(self):
        # Create a string representation of the lock
        return "<%s, %s>" % (self.__class__.__name__, "locked" if self._locked else "unlocked")


class LocalLock(FileLock):

    def __init__(self, path, temporary_directory=os.path.join(tempfile.gettempdir(), __name__)):
        # Create the directory if it does not exist
        if not os.path.isdir(temporary_directory):
            os.makedirs(temporary_directory)

        # If the path is a string, encode it
        if isinstance(path, str):
            path = path.encode(ENCODING)

        # Create hexdigest from path
        hexdigest = hashlib.md5(path).hexdigest()

        # Create the lock path based on the given path
        super(LocalLock, self).__init__(os.path.join(temporary_directory, hexdigest))


class ReferenceLock(object):

    def __init__(self, lock):
        # Initialize the internal lock
        self._lock = lock

        # Initialize the counter
        self._thread = None
        self._references = 0

    def acquire(self, blocking=True, timeout=None):
        # Fetch the current thread
        current_thread = threading.current_thread()

        # If there are no references, lock the lock
        if current_thread != self._thread:
            # Lock the lock
            self._lock.acquire(blocking=blocking, timeout=timeout)

            # Set the thread
            self._thread = current_thread

        # Increment the reference count
        self._references += 1

    def release(self):
        # Fetch the current thread
        current_thread = threading.current_thread()

        # Make sure the thread is the locking thread
        if current_thread != self._thread:
            raise RuntimeError("Thread %r can't release %r" % (current_thread, self))

        # Check the reference count
        if not self._references:
            raise RuntimeError("Already released")

        # Decrement the references
        self._references -= 1

        # Check if should release lock
        if not self._references:
            # Clear the thread
            self._thread = None

            # Release the lock
            self._lock.release()

    def __enter__(self):
        # Lock the lock
        self.acquire()

        # Return "self"
        return self

    def __exit__(self, *exc_info):
        # Unlock the lock
        self.release()

    def __str__(self):
        # Create a string representation of the lock
        return "<%s, %d references>" % (self.__class__.__name__, self._references)



import pytest
import string
import random
import tempfile
import itertools
import multiprocessing

from fsdicts import *


@pytest.fixture(params=itertools.product([PYTHON, JSON], [Dictionary, AttributeDictionary], [LinkStorage, ReferenceStorage], [LocalLock, FileLock]))
def database(request):
    return fsdict(tempfile.mktemp(), *request.param)


@pytest.fixture(params=itertools.product([PYTHON, JSON], [AttributeDictionary], [LinkStorage, ReferenceStorage], [LocalLock, FileLock]))
def bunch_database(request):
    return fsdict(tempfile.mktemp(), *request.param)


def test_write_read_has_delete(database):
    # Make sure the database does not have the item
    assert "Hello" not in database

    # Write the Hello value
    database["Hello"] = "World"

    # Read the Hello value
    assert database["Hello"] == "World"

    # Make sure the database has the Hello item
    assert "Hello" in database

    # Delete the item
    del database["Hello"]

    # Make sure the database does not have the item
    assert "Hello" not in database

    # Make sure the getter now raises
    with pytest.raises(KeyError):
        assert database["Hello"] == "World"


def test_write_recursive_dicts(database):
    # Write the Hello value
    database["Hello"] = {"World": 42}

    # Read the Hello value
    assert database["Hello"] == dict(World=42)

    # Make sure the Hello value is a dictionary
    assert isinstance(database["Hello"], Dictionary)


def test_storage_usage(database):
    # Write the Hello value
    database["Hello"] = {"World": {"Another": "Value"}}

    # Make sure the storage usage is NOT zero
    assert len(database._key_storage)

    # Clear the database
    database.clear()

    # Make sure the database is empty
    assert len(database) == 0

    # Make sure the storage usage is zero
    assert not len(database._key_storage)


def test_bunch_write_read_has_delete(bunch_database):
    # Make sure the database does not have the attribute
    assert not hasattr(bunch_database, "hello_world")

    # Write to the database
    bunch_database.hello_world = "Hello World!"

    # Read the value and validate
    assert bunch_database.hello_world == "Hello World!"

    # Make sure the database has the attribute
    assert hasattr(bunch_database, "hello_world")

    # Delete the item
    del bunch_database.hello_world

    # Make sure the database does not have the attribute
    assert not hasattr(bunch_database, "hello_world")

    # Make sure a KeyError is raised
    with pytest.raises(AttributeError):
        assert bunch_database.hello_world == "Hello World!"


def test_len(database):
    # Make sure database is empty
    assert not database

    # Load value to database
    database["Hello"] = "World"

    # Make sure database is not empty
    assert database


def test_pop(database):
    # Load value to database
    database["Hello"] = "World"

    # Pop the item from the database
    assert database.pop("Hello") == "World"

    # Make sure the database is empty
    assert not database


def test_popitem(database):
    # Load value to database
    database["Hello"] = "World"

    # Pop the item from the database
    assert database.popitem() == ("Hello", "World")

    # Make sure the database is empty
    assert not database


def test_copy(database):
    # Load values to database
    database["Hello1"] = "World1"
    database["Hello2"] = "World2"

    # Copy the database and compare
    copy = database.copy()

    # Check copy
    assert isinstance(copy, dict)
    assert copy == {"Hello1": "World1", "Hello2": "World2"}


def test_equals(database):
    # Load values to database
    database["Hello1"] = "World1"
    database["Hello2"] = "World2"

    assert database == {"Hello1": "World1", "Hello2": "World2"}
    assert database != {"Hello1": "World1", "Hello2": "World2", "Hello3": "World3"}
    assert database != {"Hello2": "World2", "Hello3": "World3"}


def test_representation(database):
    # Make sure looks good empty
    assert repr(database) == "{}"

    # Load some values
    database["Hello"] = "World"
    database["Other"] = {"Test": 1}

    # Make sure looks good with data
    assert repr(database) in ["{'Hello': 'World', 'Other': {'Test': 1}}", "{'Other': {'Test': 1}, 'Hello': 'World'}"] + ["{u'Hello': u'World', u'Other': {u'Test': 1}}", "{u'Other': {u'Test': 1}, u'Hello': u'World'}"]


def test_clear(database):
    # Load some values
    database["Hello"] = "World"
    database["Other"] = {"Test": 1}

    # Fetch other database
    other = database["Other"]

    # Make sure other is not empty
    assert other

    # Clear the database
    database.clear()

    # Make sure database is empty
    assert not database

    # Make sure other does not exist
    with pytest.raises(OSError):
        assert not other


def test_multiprocess_writes(database):
    # Create global things
    manager = multiprocessing.Manager()
    exceptions = manager.list()

    def stress():
        for _ in range(100):
            try:
                # Create random data
                rand = random.shuffle(list(string.ascii_letters))

                # Write to database
                database["item"] = rand
            except BaseException as e:
                # Append failure
                exceptions.append(e)

    # Create many stress processes
    processes = [multiprocessing.Process(target=stress) for _ in range(10)]

    # Execute all processes
    for p in processes:
        p.start()

    # Wait for all processes
    for p in processes:
        p.join()

    # Raise all of the exceptions
    for e in exceptions:
        raise e



import os

from fsdicts.lock import FileLock, LocalLock
from fsdicts.encoders import JSON, PYTHON
from fsdicts.storage import ReferenceStorage, LinkStorage
from fsdicts.dictionary import AttributeDictionary


def fsdict(path, encoder=JSON, dictionary=AttributeDictionary, storage=ReferenceStorage, lock=FileLock):
    # Create the directory
    if not os.path.exists(path):
        os.makedirs(path)

    # Initialize the storage object
    key_storage = storage(os.path.join(path, "keys"), lock=lock)
    value_storage = storage(os.path.join(path, "values"), lock=lock)

    # Initialize the keystore with objects path and a rainbow table
    return dictionary(os.path.join(path, "structure"), (key_storage, value_storage), encoder, lock)


def localdict(path, encoder=PYTHON):
    # Pick the best locks and storage for use-case
    if os.name == "posix":
        return fsdict(path, encoder=encoder, dictionary=AttributeDictionary, storage=LinkStorage, lock=LocalLock)
    else:
        return fsdict(path, encoder=encoder, dictionary=AttributeDictionary, storage=ReferenceStorage, lock=LocalLock)



import pytest

from fsdicts.bunch import Bunch


def test_bunch():
    # Create a test bunch
    bunch = Bunch(hello="World")

    # Make sure it has the attribute
    assert hasattr(bunch, "hello")
    assert not hasattr(bunch, "World")

    # Write the World attribute using setitem
    bunch["World"] = "Hello"

    # Make sure the bunch has the attribute now
    assert hasattr(bunch, "World")

    # Make sure attributes can be deleted
    del bunch.World

    # Make sure the attribute does not exist
    assert not hasattr(bunch, "World")

    # Make sure the getter raises attribute error
    with pytest.raises(AttributeError):
        a = bunch.World

    # Make sure the getitem raised key error
    with pytest.raises(KeyError):
        b = bunch["World"]


import os
import hashlib

from fsdicts.bunch import MutableAttributeMapping
from fsdicts.mapping import AdvancedMutableMapping, Mapping

# Create key and value prefixes
FILE_KEY = "key"
FILE_VALUE = "value"

# Create default object so that None can be used as default value
DEFAULT = object()


class Dictionary(AdvancedMutableMapping):

    def __init__(self, path, storage, encoder, lock):
        # Make the path absolute
        path = os.path.abspath(path)

        # Set internal lock class
        self._lock = lock

        # Set internal variables
        self._path = path
        self._encode, self._decode = encoder
        self._key_storage, self._value_storage = storage

        # Make sure path exists
        if not os.path.exists(self._path):
            os.makedirs(self._path)

    def _child_instance(self, path):
        return Dictionary(path, (self._key_storage, self._value_storage), (self._encode, self._decode), self._lock)

    def _internal_iter(self):
        # List all of the items in the path
        for checksum in os.listdir(self._path):
            # Create key path
            key_path = os.path.join(self._path, checksum, FILE_KEY)

            # Make sure the key path exists
            if not os.path.isfile(key_path):
                continue

            # Yield the checksum
            yield checksum, key_path

    def _internal_hasitem(self, key):
        # Resolve item path
        item_path = self._item_path(key)

        # Check if paths exist
        return os.path.isdir(item_path)

    def _internal_setitem(self, key, value):
        # Delete the old value
        if self._internal_hasitem(key):
            self._internal_delitem(key)

        # Resolve item path
        item_path = self._item_path(key)

        # Create the item path
        os.makedirs(item_path)

        # Resolve value path
        key_path, value_path = os.path.join(item_path, FILE_KEY), os.path.join(item_path, FILE_VALUE)

        # Store the key in the object storage
        self._key_storage.put(self._encode(key), key_path)

        # Check if value is a dictionary
        if isinstance(value, Mapping):
            # Create the sub-dictionary
            dictionary = self._child_instance(value_path)

            # Update the dictionary with the values
            dictionary.update(value)
        else:
            # Store the value in the object storage
            self._value_storage.put(self._encode(value), value_path)

    def _internal_getitem(self, key):
        # Make sure key exists
        if not self._internal_hasitem(key):
            raise KeyError(key)

        # Resolve item path
        item_path = self._item_path(key)

        # Resolve value path
        value_path = os.path.join(item_path, FILE_VALUE)

        # Check if object is a simple object
        if os.path.isdir(value_path):
            # Create a keystore from the path
            return self._child_instance(value_path)
        else:
            # Read the value, decode and return
            return self._decode(self._value_storage.readlink(value_path))

    def _internal_delitem(self, key):
        # Make sure key exists
        if not self._internal_hasitem(key):
            raise KeyError(key)

        # Resolve item path
        item_path = self._item_path(key)

        # Resolve value path
        key_path, value_path = os.path.join(item_path, FILE_KEY), os.path.join(item_path, FILE_VALUE)

        # Unlink the key object
        self._key_storage.unlink(key_path)

        # If the value is a dictionary, clear the dictionary
        if os.path.isdir(value_path):
            # Create the dictionary
            dictionary = self._child_instance(value_path)

            # Clear the dictionary
            dictionary.clear()

            # Delete the directory - should be empty
            os.rmdir(value_path)
        else:
            # Unlink the value object
            self._value_storage.unlink(value_path)

        # Remove the item path
        os.rmdir(item_path)

    def _item_path(self, key):
        # Hash the key using hashlib
        checksum = hashlib.md5(self._encode(key)).hexdigest()

        # Create both paths
        return os.path.join(self._path, checksum)

    def __getitem__(self, key):
        # Resolve item path
        item_path = self._item_path(key)

        # Lock the item for modifications
        with self._lock(item_path):
            return self._internal_getitem(key)

    def __setitem__(self, key, value):
        # Resolve item path
        item_path = self._item_path(key)

        # Lock the item for modifications
        with self._lock(item_path):
            return self._internal_setitem(key, value)

    def __delitem__(self, key):
        # Resolve item path
        item_path = self._item_path(key)

        # Lock the item for modifications
        with self._lock(item_path):
            return self._internal_delitem(key)

    def __contains__(self, key):
        # Resolve item path
        item_path = self._item_path(key)

        # Check if paths exist
        with self._lock(item_path):
            return self._internal_hasitem(key)

    def __iter__(self):
        # List all of the items in the path
        for checksum, key_path in self._internal_iter():
            # Read the key contents and decode
            with self._lock(os.path.join(self._path, checksum)):
                key_contents = self._decode(self._key_storage.readlink(key_path))

            # Yield the key contents
            yield key_contents

    def __len__(self):
        # Count all key files
        return len(list(self._internal_iter()))

    def __repr__(self):
        # Format the data like a dictionary
        return "{%s}" % ", ".join("%r: %r" % item for item in self.items())

    def __eq__(self, other):
        # Make sure the other object is a mapping
        if not isinstance(other, Mapping):
            return False

        # Make sure all keys exist
        if set(self.keys()) != set(other.keys()):
            return False

        # Make sure all the values equal
        for key in self:
            if self[key] != other[key]:
                return False

        # Comparison succeeded
        return True

    def pop(self, key, default=DEFAULT):
        try:
            # Resolve item path
            item_path = self._item_path(key)

            # Check if paths exist
            with self._lock(item_path):
                # Fetch the value
                value = self._internal_getitem(key)

                # Check if the value is a keystore
                if isinstance(value, Mapping):
                    value = value.copy()

                # Delete the item
                self._internal_delitem(key)

            # Return the value
            return value
        except KeyError:
            # Check if a default is defined
            if default != DEFAULT:
                return default

            # Reraise exception
            raise

    def popitem(self):
        # Convert self to list
        keys = list(self)

        # If the list is empty, raise
        if not keys:
            raise KeyError()

        # Pop a key from the list
        key = keys.pop()

        # Return the key and the value
        return key, self.pop(key)

    def copy(self):
        # Create initial bunch
        output = dict()

        # Loop over keys
        for key in self:
            # Fetch value of key
            value = self[key]

            # Check if value is a keystore
            if isinstance(value, Mapping):
                value = value.copy()

            # Update the bunch
            output[key] = value

        # Return the created output
        return output

    def clear(self):
        # Loop over keys
        for key in self:
            # Delete the item
            del self[key]


class AttributeDictionary(Dictionary, MutableAttributeMapping):

    _lock = None
    _path = None
    _encode, _decode = None, None
    _key_storage, _value_storage = None, None

    def _child_instance(self, path):
        return AttributeDictionary(path, (self._key_storage, self._value_storage), (self._encode, self._decode), self._lock)



import json
import functools

try:
    # Python 3 mapping
    from collections.abc import MutableMapping, Mapping
except:
    # Python 2 mapping
    from collections import MutableMapping, Mapping


class AdvancedMutableMapping(MutableMapping):

    def setdefaults(self, *dictionaries, **values):
        # Update values to include all dicts
        for dictionary in dictionaries:
            values.update(dictionary)

        # Loop over all items and set the default value
        for key, value in values.items():
            self.setdefault(key, value)


class MappingEncoder(json.JSONEncoder):

    def default(self, obj):
        # Check if the object is a keystore
        if isinstance(obj, Mapping):
            # Return a JSON encodable representation of the keystore
            return obj.copy()

        # Fallback default
        return super(MappingEncoder, self).default(obj)


# Update the default dumps function
json.dump = functools.partial(json.dump, cls=MappingEncoder)
json.dumps = functools.partial(json.dumps, cls=MappingEncoder)



import os
import time
import tempfile
import threading

from fsdicts import *


def test_lock():
    # Create path to lock on
    path = tempfile.mktemp()

    # FileLock the path
    with FileLock(path) as lock:
        assert os.path.isdir(lock._path)


def test_lock_multithreaded(num_threads=5, thread_sleep=0.2):
    # Create path to lock on
    path = tempfile.mktemp()

    def target(path, sleep):
        # Try locking the path
        with FileLock(path):
            time.sleep(sleep)

    # Create threads
    threads = [threading.Thread(target=target, args=(path, thread_sleep)) for _ in range(num_threads)]

    # Mark start time
    start = time.time()

    # Start all threads
    for t in threads:
        t.start()

    for t in threads:
        t.join()

    # Make sure end time is larger then start time by more then num_threads * thread_sleep
    assert (time.time() - start) > float(num_threads * thread_sleep)


def test_lock_multithreaded_samelock(num_threads=5, thread_sleep=0.2):
    # Create path to lock on
    path = tempfile.mktemp()
    lock = FileLock(path)

    def target(lock, sleep):
        # Try locking the path
        with lock:
            time.sleep(sleep)

    # Create threads
    threads = [threading.Thread(target=target, args=(lock, thread_sleep)) for _ in range(num_threads)]

    # Mark start time
    start = time.time()

    # Start all threads
    for t in threads:
        t.start()

    for t in threads:
        t.join()

    # Make sure end time is larger then start time by more then num_threads * thread_sleep
    assert (time.time() - start) > float(num_threads * thread_sleep)


def test_lock_nonblocking():
    # Create path to lock on
    path = tempfile.mktemp()

    # Create the lock
    lock = FileLock(path)

    # Try locking the lock
    assert lock.acquire(False)
    assert not lock.acquire(False)

    # Release the lock
    lock.release()

    # Acquire the lock
    assert lock.acquire()

    # Mark start time
    start_time = time.time()

    # Try aquiring the lock
    assert not lock.acquire(timeout=1)

    # Check end time
    assert time.time() - start_time > 1


def test_rlock_references():
    # Create path to lock on
    path = tempfile.mktemp()

    # Create the lock
    lock = FileLock(path)
    mutex = ReferenceLock(lock)

    # FileLock the lock multiple times
    with mutex:
        with mutex:
            with mutex:
                assert mutex._references == 3

    # Check empty references
    assert mutex._references == 0


def test_rlock_multithreaded_samelock(num_threads=5, thread_sleep=0.2):
    # Create path to lock on
    path = tempfile.mktemp()

    # Create the lock
    lock = FileLock(path)
    mutex = ReferenceLock(lock)

    def target(mutex, number):
        # Try locking the path
        with mutex:
            with mutex:
                time.sleep(number)

    # Create threads
    threads = [threading.Thread(target=target, args=(mutex, thread_sleep)) for _ in range(num_threads)]

    # Mark start time
    start = time.time()

    # Start all threads
    for t in threads:
        t.start()

    for t in threads:
        t.join()

    # Make sure end time is larger then start time by more then num_threads * thread_sleep
    assert (time.time() - start) > float(num_threads * thread_sleep)



import json
import pytest

from fsdicts import *


def test_setdefaults():
    # Create test type
    class AdvancedBunch(Bunch, AdvancedMutableMapping):
        pass

    # Create a bunch
    bunch = AdvancedBunch(hello="World")
    bunch.setdefaults(hello="Changed!", world="Good")

    # Check values
    assert bunch.hello == "World"
    assert bunch.world == "Good"


def test_json_encoding():
    # Create a test class that imitates a mapping
    class MyMapping(Mapping):

        def __getitem__(self, key):
            return "World"

        def __len__(self):
            return 1

        def __iter__(self):
            return iter(["Hello"])

        def copy(self):
            return {key: self[key] for key in self}

    # Create a test object
    my_mapping = MyMapping()

    # Try encoding the mapping
    assert json.dumps(my_mapping) == '{"Hello": "World"}'



# Encoder tuples
from fsdicts.encoders import JSON, PYTHON

# Constructor functions
from fsdicts.database import fsdict, localdict

# Underlaying classes
from fsdicts.lock import FileLock, LocalLock, ReferenceLock
from fsdicts.bunch import AttributeMapping, Bunch
from fsdicts.mapping import AdvancedMutableMapping, MutableMapping, Mapping
from fsdicts.storage import Storage, LinkStorage, ReferenceStorage
from fsdicts.dictionary import Dictionary, AttributeDictionary